import torch
import numpy as np
from einops import rearrange
from ABlooper.models import DecoyGen
from ABlooper.utils import filt, rmsd, to_pdb_line, short2long, long2short, prepare_input_loop
import os

try:
    from ABlooper.rosetta_refine import rosetta_refine
except ModuleNotFoundError:
    rosetta_available = False
else:
    rosetta_available = True

try:
    from ABlooper.openmm_refine import openmm_refine
except ModuleNotFoundError:
    openmm_available = False
else:
    openmm_available = True

device = "cuda" if torch.cuda.is_available() else "cpu"
current_directory = os.path.dirname(os.path.realpath(__file__))

default_model = DecoyGen().float().to(device)
path = os.path.join(current_directory, "trained_model", "final_model")
default_model.load_state_dict(torch.load(path, map_location=torch.device(device)))


class CDR_Predictor:
    def __init__(self, pdb_file, chains=("H", "L"), model=None, refine=False, refine_method="openmm"):
        """ Class used to handle remodelling of CDRs.

        :param pdb_file: File of IMGT numbered antibody structure in .pdb format. Must include heavy and light chain
        :param chains: Name of heavy and light chain to be remodelled in the file
        :param model: Trained neural network. If left as None a pretrained network is used.
        """
        self.pdb_file = pdb_file
        self.chains = chains
        self.model = model if model is not None else default_model
        self.CDR_with_anchor_slices = {
            "H1": (chains[0], (25, 39)),
            "H2": (chains[0], (55, 66)),
            "H3": (chains[0], (105, 119)),
            "L1": (chains[1], (22, 42)),
            "L2": (chains[1], (54, 71)),
            "L3": (chains[1], (103, 119))}
        self.__atoms = ["CA", "N", "C", "CB"]

        self.predicted_CDRs = {}
        self.all_decoys = {}
        self.decoy_diversity = {}

        with open(self.pdb_file) as file:
            self.pdb_text = [line for line in file.readlines()]

        # For all three of these I extract the loop plus two anchors at either side as these are needed for the model.
        self.CDR_text = {CDR: [x for x in self.pdb_text if filt(x, *self.CDR_with_anchor_slices[CDR])] for CDR in
                         self.CDR_with_anchor_slices}
        self.CDR_sequences = {
            CDR: "".join([long2short[x.split()[3][-3:]] for x in self.CDR_text[CDR] if x.split()[2] == "CA"]) for CDR in
            self.CDR_with_anchor_slices}
        self.__extract_BB_coords()

        # Here I don't extract the anchors as this is only needed for writing predictions to pdb file.
        self.CDR_numberings = {CDR: [x.split()[5] for x in self.CDR_text[CDR] if x.split()[2] == "CA"][2:-2] for CDR in
                               self.CDR_text}
        self.CDR_start_atom_id = {CDR: int([x.split()[1] for x in self.CDR_text[CDR] if x.split()[2] == "N"][2]) for CDR
                                  in self.CDR_text}

        # If rosetta or openmm are not available refinement is not possible
        if (not rosetta_available) and (not openmm_available) and refine:
            print("Neither PyRosetta nor OpenMM are available. Refinement is not possible")
        self.refine = refine and (rosetta_available or openmm_available)
        self.refine_method = refine_method

    def __extract_BB_coords(self):
        self.CDR_BB_coords = {}

        for CDR in self.CDR_with_anchor_slices:
            loop = self.CDR_text[CDR]

            coors = np.zeros((len(self.CDR_sequences[CDR]), 4, 3))
            coors[...] = float("Nan")

            i = 0
            res = loop[i].split()[5]

            for line in loop:
                cut = line.split()
                if cut[5] != res:
                    res = cut[5]
                    i += 1
                if cut[2] in self.__atoms:
                    j = self.__atoms.index(cut[2])
                    # Using split for corrds doesn't always work. Following Biopython approach:
                    x = float(line[30:38])
                    y = float(line[38:46])
                    z = float(line[46:54])
                    coors[i, j] = np.array([x, y, z])

            # If missed CB (GLY) then add CA instead
            coors[:, 3] = np.where(np.all(coors[:, 3] != coors[:, 3], axis=-1, keepdims=True), coors[:, 0], coors[:, 3])
            self.CDR_BB_coords[CDR] = coors

    def __prepare_model_input(self):
        encodings = []
        geomins = []

        for CDR in self.CDR_BB_coords:
            geom, encode = prepare_input_loop(self.CDR_BB_coords[CDR], self.CDR_sequences[CDR], CDR)
            encodings.append(encode)
            geomins.append(geom)

        return torch.cat(encodings, axis=1).detach(), torch.cat(geomins, axis=1).detach()

    def predict_CDRs(self):

        with torch.no_grad():
            input_encoding, input_geometry = self.__prepare_model_input()
            output = self.model(input_encoding.to(device), input_geometry.to(device))

            for i, CDR in enumerate(self.CDR_with_anchor_slices):
                output_CDR = output[:, input_encoding[0, :, 30 + i] == 1.0]
                self.all_decoys[CDR] = rearrange(output_CDR, "b (i a) d -> b i a d", a=4).cpu().numpy()
                self.predicted_CDRs[CDR] = rearrange(output_CDR.mean(0), "(i a) d -> i a d", a=4).cpu().numpy()
                self.decoy_diversity[CDR] = (output_CDR[None] - output_CDR[:, None]).pow(2).sum(-1).mean(-1).pow(
                    1 / 2).sum().item() / 20

        return self.predicted_CDRs

    def calculate_BB_rmsd_wrt_input(self):
        assert "H1" in self.predicted_CDRs, "Predict it before comparing"
        rmsds = {}

        for CDR in self.CDR_BB_coords:
            rmsds[CDR] = rmsd(self.CDR_BB_coords[CDR][2:-2], self.predicted_CDRs[CDR])

        return rmsds

    def __convert_predictions_into_text_for_each_CDR(self):
        pdb_format = {}

        pdb_atoms = ["N", "CA", "C", "CB"]
        permutation_to_reorder_atoms = [1, 0, 2, 3]

        for CDR in self.CDR_start_atom_id:
            new_text = []
            BB_coords = self.predicted_CDRs[CDR]
            seq = self.CDR_sequences[CDR][2:-2]
            numbering = self.CDR_numberings[CDR]
            atom_id = self.CDR_start_atom_id[CDR]
            chain = self.CDR_with_anchor_slices[CDR][0]

            for i, amino in enumerate(BB_coords):
                amino_type = short2long[seq[i]]
                for j, coord in enumerate(amino[permutation_to_reorder_atoms]):
                    if (pdb_atoms[j] == "CB") and (amino_type == "GLY"):
                        continue
                    new_text.append(to_pdb_line(atom_id, pdb_atoms[j], amino_type, chain, numbering[i], coord))
                    atom_id += 1

            pdb_format[CDR] = new_text
        return pdb_format

    def write_predictions_in_pdb_format(self, file_name=None, to_be_rewritten=None):
        if to_be_rewritten is None:
            # This should probably be left as is to avoid clashes between loops
            to_be_rewritten = ["H1", "H2", "H3", "L1", "L2", "L3"]
        if "H1" not in self.predicted_CDRs:
            self.predict_CDRs()

        text_prediction_per_CDR = self.__convert_predictions_into_text_for_each_CDR()
        old_text = self.pdb_text

        for CDR in to_be_rewritten:
            new = True
            new_text = []
            chain, CDR_slice = self.CDR_with_anchor_slices[CDR]
            CDR_slice = (CDR_slice[0] + 2, CDR_slice[1] - 2)

            for line in old_text:
                if not filt(line, chain, CDR_slice):
                    new_text.append(line)
                elif new:
                    new_text += text_prediction_per_CDR[CDR]
                    new = False
                else:
                    continue
            old_text = new_text

        header = [
            "REMARK    CDR LOOPS REMODELLED USING ABLOOPER                                   \n"]

        if self.refine:
            if self.refine_method == "openmm" and openmm_available:
                old_text = openmm_refine(old_text, self.CDR_with_anchor_slices)
                header.append("REMARK    REFINEMENT DONE USING OPENMM" + 42 * " " + "\n")
            else:
                old_text = rosetta_refine(old_text, self.CDR_with_anchor_slices)
                header.append("REMARK    REFINEMENT DONE USING PYROSETTA" + 39 * " " + "\n")

        new_text = "".join(header + old_text)

        if file_name is None:
            return "".join(new_text)
        else:
            with open(file_name, "w+") as file:
                file.write(new_text)

    def __repr__(self):
        return "CDR_Predictor: {}".format(self.pdb_file.split("/")[-1])



import argparse
from ABlooper.ABlooper import CDR_Predictor

try:
    from ABlooper.openmm_refine import openmm_refine
except ModuleNotFoundError:
    try:
        from ABlooper.rosetta_refine import rosetta_refine
    except ModuleNotFoundError:
        refinement_available = False
    else:
        refinement_available = True
else:
    refinement_available = True

parser = argparse.ArgumentParser()
parser.add_argument("file_path", help="Path to the IMGT numbered antibody pdb file for which the CDRs are to be "
                                      "remodelled")
parser.add_argument("-o", "--output", help="Path to where the output model should be saved. Defaults to the same "
                                           "directory as input file.", default=None)
parser.add_argument("-H", "--heavy_chain", help="Heavy chain ID for input file. (Default is H)", default="H")
parser.add_argument("-L", "--light_chain", help="Light chain ID for input file. (Default is L)", default="L")
parser.add_argument("--confidence_score", help="Print confidence score for each loop", default=False,
                    action="store_true")
if refinement_available:
    parser.add_argument("-s", "--side_chains", help="Predict side chains and refine loop geometry",
                        default=False, action="store_true")

args = parser.parse_args()


def main():
    if refinement_available:
        side_chains = args.side_chains
    else:
        side_chains = False

    predictor = CDR_Predictor(args.file_path, chains=(args.heavy_chain, args.light_chain), refine=side_chains)
    output_file = args.output

    if output_file is None:
        output_file = args.file_path[:-4] + "_remodelled_CDRs.pdb"

    predictor.write_predictions_in_pdb_format(output_file)

    if args.confidence_score is True:
        print(predictor.decoy_diversity)



import pyrosetta
import tempfile
from ABlooper.utils import stop_print

stop_print(pyrosetta.init, "-mute all")

# The code in this file is an adaptation of scripts written by Jeffrey Ruffolo (jeffreyruffolo).
# Many thanks to him for his help


def get_loop_ranges(chain, res, info):
    """ Find loops in PyRosetta pose numbering

    """
    start = info.pdb2pose(chain, res[0])
    end = info.pdb2pose(chain, res[1])

    # Basically if you can't find the start or end residues, try the next residue until you run out.
    if start == 0:
        i = res[0]
        while (start == 0) and (i > 18):
            i -= 1
            start = info.pdb2pose(chain, i)

    if end == 0:
        i = res[1]
        while (end == 0) and (i < 122):
            i += 1
            end = info.pdb2pose(chain, i)

    return start, end


def get_fa_min_mover(loop_ranges, pose, min_iter):
    """ Create full-atom minimization mover

    """
    # Create full-atom score function with terms for fixing bad bond lengths
    sf = pyrosetta.create_score_function('ref2015_cst')
    sf.set_weight(pyrosetta.rosetta.core.scoring.ScoreType.cart_bonded, 1)
    sf.set_weight(pyrosetta.rosetta.core.scoring.ScoreType.pro_close, 0)

    # Try to minimize only the loops
    try:
        loops = pyrosetta.rosetta.protocols.loops.Loops()
        [loops.add_loop(r[0], r[1]) for r in loop_ranges]

        mmap = pyrosetta.rosetta.protocols.loops.move_map_from_loops(
            pose, loops, False)

    # If selecting the loops doesn't work, minimise the whole thing
    except RuntimeError:
        mmap = pyrosetta.rosetta.core.kinematics.MoveMap()
        mmap.set_bb(True)
        mmap.set_chi(True)
        mmap.set_jump(True)

    # Create MinMover acting in cartesian space
    min_mover = pyrosetta.rosetta.protocols.minimization_packing.MinMover(
        mmap, sf, 'lbfgs_armijo_nonmonotone', 0.0001, True)
    min_mover.max_iter(min_iter)
    min_mover.cartesian(True)

    return min_mover


def rosetta_refine(pdb_txt, CDR_definitions, min_iter=500):
    """ PyRosetta protocol for minimization refinement of protein structure

    """
    # Load input PDB into pose
    with tempfile.NamedTemporaryFile("wt") as tmp:
        tmp.writelines(pdb_txt)
        pose = pyrosetta.pose_from_pdb(tmp.name)

    # Get PDB info
    info = pose.pdb_info()

    # Define CDR loops
    loop_ranges = [get_loop_ranges(*x, info) for x in CDR_definitions.values()]

    # Create movers
    cst_mover = pyrosetta.rosetta.protocols.relax.AtomCoordinateCstMover()
    cst_mover.cst_sidechain(False)
    min_mover = get_fa_min_mover(
        loop_ranges=loop_ranges,
        pose=pose,
        min_iter=min_iter,
    )
    idealize_mover = pyrosetta.rosetta.protocols.idealize.IdealizeMover()

    # Refine structure
    cst_mover.apply(pose)
    min_mover.apply(pose)
    idealize_mover.apply(pose)

    # Save refined structure to string
    with tempfile.NamedTemporaryFile("rt") as tmp:
        pose.dump_pdb(tmp.name)
        txt = tmp.readlines()

    # Remove Rosetta details
    return [x for x in txt if x[:3] in ["ATO", "TER"]]
